from tests.globals.documents.datetime_documents import *
from tests.globals.documents.error_documents import *
from tests.globals.documents.misc_documents import *
from tests.globals.documents.nested_documents import *
from tests.globals.documents.numpy_documents import *
from tests.globals.documents.pandas_documents import *
from tests.globals.documents.simple_documents import *
from tests.globals.documents.vector_documents import *
