from tests.globals.document.error_document import *
from tests.globals.document.simple_document import *
from tests.globals.document.datetime_document import *
from tests.globals.document.dataclass_document import *
from tests.globals.document.nested_document import *
from tests.globals.document.numpy_document import *
from tests.globals.document.pandas_document import *
from tests.globals.document.vector_document import *
