SEARCH_APP_LINK = "https://cloud.relevance.ai/dataset/{}/deploy/recent/search"
PROJECTOR_APP_LINK = "https://cloud.relevance.ai/dataset/{}/deploy/recent/projector"
CLUSTER_APP_LINK = "https://cloud.relevance.ai/dataset/{}/deploy/recent/cluster/"

SIGNUP_URL = "https://cloud.relevance.ai/sdk/api"
AUSTRALIA_URL = "https://gateway-api-aueast.relevance.ai/latest"
WIDER_URL = "https://api.{}.relevance.ai/latest"
DEV_URL = "https://api-dev.{}.relevance.ai/latest"
